from .console import Console as Console, LogBackends as LogBackends, LogExtensions as LogExtensions, LogLevel as LogLevel

__all__ = ['console', 'Console', 'LogLevel', 'LogBackends', 'LogExtensions']

console: Console



"""This library exposes a minimalistic set of shared utility functions used to support other projects.

This library has a very broad scope: it centralizes widely shared functionality used across multiple
Sun Lab projects. Any function reused by more than 5 other projects becomes a candidate for inclusion into this library.

Currently, the library provides the following functionality:
- Console: A class that provides message and error printing and logging functionality.

While this library is explicitly configured to work with other Sun Lab projects, it can be adapted to work for non-lab
projects. Specifically, this would likely require changing default argument values used by functions exposed through
this library.
"""

from .console import Console, LogLevel, LogBackends, LogExtensions

# Preconfigures and exposes Console class instance as a variable, similar to how Loguru exposes logger. This instance
# can be used globally instead of defining a custom console variable.
console: Console = Console(logger_backend=LogBackends.LOGURU)

__all__ = ["console", "Console", "LogLevel", "LogBackends", "LogExtensions"]



"""This module stores the tests for all classes and functions available from the console.py module."""

import os
import re
import sys
from typing import Any, Generator
from pathlib import Path
import tempfile
import textwrap

import pytest
from pydantic import ValidationError

from ataraxis_base_utilities import Console, LogLevel, LogBackends, LogExtensions


@pytest.fixture
def temp_dir() -> Generator[Path, Any, None]:
    """Generates and yields the temporary directory used by the tests that involve log file operations."""
    tmpdirname: str
    with tempfile.TemporaryDirectory() as tmpdirname:
        yield Path(tmpdirname)


def error_format(message: str) -> str:
    """Formats the input message to match the default Console format and escapes it using re, so that it can be used to
    verify raised exceptions.

    This method is used to set up pytest 'match' clauses to verify raised exceptions.
    """
    return re.escape(textwrap.fill(message, width=120, break_long_words=False, break_on_hyphens=False))


@pytest.mark.parametrize("backend", [LogBackends.LOGURU, LogBackends.CLICK])
def test_console_initialization(backend, temp_dir) -> None:
    """Tests successful console initialization."""
    console = Console(
        logger_backend=backend,
        message_log_path=temp_dir / f"message{LogExtensions.LOG}",
        error_log_path=temp_dir / f"error{LogExtensions.TXT}",
        debug_log_path=temp_dir / f"debug{LogExtensions.JSON}",
    )
    assert console._backend == backend
    assert console._message_log_path == temp_dir / "message.log"
    assert console._error_log_path == temp_dir / "error.txt"
    assert console._debug_log_path == temp_dir / "debug.json"


@pytest.mark.parametrize("backend", [LogBackends.LOGURU, LogBackends.CLICK])
def test_console_invalid_initialization_argument_type(backend, temp_dir) -> None:
    """Tests that pydantic wrapper successfully catches and handles invalid initialization argument types.

    Also verifies that pydantic attempts to convert valid equivalents to the correct type, eg: int 1 -> bool True.
    """
    with pytest.raises(ValidationError):
        # noinspection PyTypeChecker
        Console(logger_backend=None)
    with pytest.raises(ValidationError):
        # noinspection PyTypeChecker
        Console(message_log_path=123)
    with pytest.raises(ValidationError):
        # noinspection PyTypeChecker
        Console(line_width="None")
    with pytest.raises(ValidationError):
        # noinspection PyTypeChecker
        Console(break_on_hyphens=None)
    # noinspection PyTypeChecker
    Console(break_on_hyphens=1)  # Works due to pydantic automatically parsing bool-equivalents as bools.
    with pytest.raises(ValidationError):
        # noinspection PyTypeChecker
        Console(use_color=None)


@pytest.mark.parametrize("backend", [LogBackends.LOGURU, LogBackends.CLICK])
def test_console_invalid_initialization_line_width(backend) -> None:
    """Tests invalid line_width input during Console initialization."""

    # Uses an invalid width of <= 0
    message = (
        f"Invalid 'line_width' argument encountered when instantiating Console class instance. "
        f"Expected a value greater than 0, but encountered {0}."
    )
    with pytest.raises(ValueError, match=error_format(message)):
        Console(logger_backend=backend, line_width=0)


@pytest.mark.parametrize("backend", [LogBackends.LOGURU, LogBackends.CLICK])
def test_console_invalid_initialization_log_paths(backend, temp_dir) -> None:
    """Tests invalid path inputs during Console initialization."""
    valid_extensions: tuple[str, ...] = LogExtensions.values()

    # Uses a non-supported 'zipp' extension to trigger ValueErrors.
    message = (
        f"Invalid 'debug_log_path' argument encountered when instantiating Console class instance. "
        f"Expected a path ending in a file name with one of the supported extensions:"
        f"{', '.join(valid_extensions)}, but encountered {temp_dir / 'invalid.zipp'}."
    )
    with pytest.raises(ValueError, match=error_format(message)):
        Console(logger_backend=backend, debug_log_path=temp_dir / "invalid.zipp")
    message = (
        f"Invalid 'message_log_path' argument encountered when instantiating Console class instance. "
        f"Expected a path ending in a file name with one of the supported extensions:"
        f"{', '.join(valid_extensions)}, but encountered {temp_dir / 'invalid.zipp'}."
    )
    with pytest.raises(ValueError, match=error_format(message)):
        Console(logger_backend=backend, message_log_path=temp_dir / "invalid.zipp")
    message = (
        f"Invalid 'error_log_path' argument encountered when instantiating Console class instance. "
        f"Expected a path ending in a file name with one of the supported extensions:"
        f"{', '.join(valid_extensions)}, but encountered {temp_dir / 'invalid.zipp'}."
    )
    with pytest.raises(ValueError, match=error_format(message)):
        Console(logger_backend=backend, error_log_path=temp_dir / "invalid.zipp")


@pytest.mark.parametrize("backend", [LogBackends.LOGURU, LogBackends.CLICK])
def test_console_add_handles(backend, tmp_path, capsys) -> None:
    """Verifies that add_handles method works as expected for all supported backends."""
    # Setup
    debug_log = tmp_path / "debug.log"
    message_log = tmp_path / "message.log"
    error_log = tmp_path / "error.log"
    console = Console(
        logger_backend=backend, debug_log_path=debug_log, message_log_path=message_log, error_log_path=error_log
    )

    # Tests LOGURU backend
    if backend == LogBackends.LOGURU:
        # Removes existing handlers
        console._logger.remove()

        # Tests default behavior
        console.add_handles()
        assert len(console._logger._core.handlers) == 2  # Only message_terminal by default

        # Removes handlers for next test
        console._logger.remove()

        # Tests with all handlers
        console.add_handles(
            debug_terminal=True,
            debug_file=True,
            message_terminal=True,
            message_file=True,
            error_terminal=True,
            error_file=True,
        )
        assert len(console._logger._core.handlers) == 6

        # Tests each handler
        logger = console._logger.bind(ataraxis_terminal=True, ataraxis_log=True)
        logger.debug("Debug message")
        logger.info("Info message")
        logger.warning("Warning message")
        logger.error("Error message")

        captured = capsys.readouterr()

        # Checks terminal output
        assert "Debug message" in captured.out
        assert "Info message" in captured.out
        assert "Warning message" in captured.out
        assert "Error message" in captured.err

        # Checks file output
        debug_log_content = debug_log.read_text()
        message_log_content = message_log.read_text()
        error_log_content = error_log.read_text()

        assert "Debug message" in debug_log_content
        assert "Info message" in message_log_content
        assert "Warning message" in message_log_content
        assert "Error message" in error_log_content

        # Tests removing handlers
        console._logger.remove()
        assert len(console._logger._core.handlers) == 0

    # Tests CLICK backend
    elif backend == LogBackends.CLICK:
        # For CLICK backend, add_handles should do nothing
        initial_handlers = len(console._logger._core.handlers) if console._logger else 0
        console.add_handles()
        assert len(console._logger._core.handlers) if console._logger else 0 == initial_handlers

    # Tests has_handles property. Should be 0 for both backends, as loguru tests involve removing all handles and
    # click backend does not instantiate handles in the first place.
    assert not console.has_handles

    # Enables the console and adds handles for the tests below to work for both backends
    console.enable()
    console.add_handles(
        debug_terminal=True,
        debug_file=True,
        message_terminal=True,
        message_file=True,
        error_terminal=True,
        error_file=True,
    )

    # Tests echo method for both backends
    console.echo("Test debug", LogLevel.DEBUG, terminal=True, log=True)
    console.echo("Test message", LogLevel.INFO, terminal=True, log=True)
    console.echo("Test error", LogLevel.ERROR, terminal=True, log=True)

    captured = capsys.readouterr()

    # Checks terminal for both backends
    assert "Test debug" in captured.out
    assert "Test message" in captured.out
    assert "Test error" in captured.err

    # Checks log files for both backends
    assert "Test debug" in debug_log.read_text()
    assert "Test message" in message_log.read_text()
    assert "Test error" in error_log.read_text()


def test_loguru_specific_functionality() -> None:
    """Verifies initialization functionality specific to loguru backend."""
    console = Console(logger_backend=LogBackends.LOGURU)
    assert console._logger is not None


def test_click_specific_functionality(capsys) -> None:
    """Verifies initialization functionality specific to click backend."""

    # Console works without adding any handles
    console = Console(logger_backend=LogBackends.CLICK)
    console.enable()
    console.echo("Click message", LogLevel.INFO)
    captured = capsys.readouterr()
    assert "Click message" in captured.out

    # Console does not have a logger instance.
    assert console._logger is None


@pytest.mark.parametrize("backend", [LogBackends.LOGURU, LogBackends.CLICK])
def test_console_enable_disable(backend) -> None:
    """Tests the functionality of the console enable / disable methods and the is_enabled property."""

    # tests enable / disable methods and is_enabled tracker
    console = Console(logger_backend=backend)
    assert not console.is_enabled
    console.enable()
    assert console.is_enabled
    console.disable()
    assert not console.is_enabled

    # Verifies that echo does not process input messages when the console is disabled
    assert not console.echo(message="Test", level=LogLevel.INFO)


@pytest.mark.parametrize("backend", [LogBackends.LOGURU, LogBackends.CLICK])
def test_console_format_message(backend) -> None:
    """Verifies that loguru and non-loguru message formatting works as expected for all backends."""
    console = Console(logger_backend=backend, line_width=120)
    message = "This is a long message that should be wrapped properly according to the specified parameters"

    # Tests non-loguru wrapping
    formatted = console.format_message(message, loguru=False)
    assert len(max(formatted.split("\n"), key=len)) <= 120

    # Tests loguru wrapping
    formatted = console.format_message(message, loguru=True)
    lines = formatted.split("\n")

    # Checks first line (should be 83 characters or fewer due to 37-character loguru header)
    assert len(lines[0]) <= 83  # 120 - 37 = 83

    # Checks the following lines (should be 120 characters or fewer, with at least 37 characters of indentation)
    for line in lines[1:]:
        assert len(line) <= 120
        assert line.startswith(" " * 37)

    # Ensures that the wrapped message is not empty
    assert formatted.strip() != ""

    # Ensures that the entire original message is contained within the formatted version
    # We'll use a more flexible comparison that ignores spaces and newlines
    formatted_content = re.sub(r"\s+", "", formatted)
    message_content = re.sub(r"\s+", "", message)
    assert message_content in formatted_content

    # Additional check to ensure all words are present in the correct order
    formatted_words = re.findall(r"\w+", formatted)
    message_words = re.findall(r"\w+", message)
    assert formatted_words == message_words


@pytest.mark.parametrize("backend", [LogBackends.LOGURU, LogBackends.CLICK])
def test_console_format_message_errors(backend) -> None:
    """Verifies the format_message() error-handling functionality."""
    console = Console(logger_backend=backend, line_width=120)
    with pytest.raises(ValidationError):
        # noinspection PyTypeChecker
        console.format_message(message=123)
    with pytest.raises(ValidationError):
        # noinspection PyTypeChecker,PyArgumentList
        console.format_message(loguru=None)


@pytest.mark.parametrize("backend", [LogBackends.LOGURU, LogBackends.CLICK])
def test_console_echo(backend, tmp_path, capsys):
    # Setup
    debug_log = tmp_path / "debug.log"
    message_log = tmp_path / "message.log"
    error_log = tmp_path / "error.log"

    console = Console(
        logger_backend=backend, debug_log_path=debug_log, message_log_path=message_log, error_log_path=error_log
    )
    console.enable()
    console.add_handles(
        debug_terminal=True,
        debug_file=True,
        message_terminal=True,
        message_file=True,
        error_terminal=True,
        error_file=True,
    )

    # Tests each log level
    log_levels = [
        (LogLevel.DEBUG, debug_log),
        (LogLevel.INFO, message_log),
        (LogLevel.SUCCESS, message_log),
        (LogLevel.WARNING, message_log),
        (LogLevel.ERROR, error_log),
        (LogLevel.CRITICAL, error_log),
    ]

    for level, log_file in log_levels:
        message = f"Test {level.name} message"
        result = console.echo(message, level, terminal=True, log=True)

        assert result is True  # echo should return True when successful

        captured = capsys.readouterr()

        if level in [LogLevel.ERROR, LogLevel.CRITICAL]:
            assert message in captured.err
        else:
            assert message in captured.out

        assert os.path.exists(log_file)
        with open(log_file, "r") as f:
            assert message in f.read()

    # Tests terminal-only output
    result = console.echo("Terminal only", LogLevel.INFO, terminal=True, log=False)
    assert result is True
    captured = capsys.readouterr()
    assert "Terminal only" in captured.out
    with open(message_log, "r") as f:
        assert "Terminal only" not in f.read()

    # Tests log-only output
    result = console.echo("Log only", LogLevel.INFO, terminal=False, log=True)
    assert result is True
    captured = capsys.readouterr()
    assert "Log only" not in captured.out
    with open(message_log, "r") as f:
        assert "Log only" in f.read()

    # Tests when disabled console behavior
    console.disable()
    result = console.echo("Disabled message", LogLevel.INFO, terminal=True, log=True)
    assert result is False
    captured = capsys.readouterr()
    assert "Disabled message" not in captured.out
    with open(message_log, "r") as f:
        assert "Disabled message" not in f.read()

    # Tests with a very long message
    long_message = "This is a very long message " * 20
    result = console.echo(long_message, LogLevel.INFO, terminal=True, log=True)
    assert result is False  # Because console is still disabled
    captured = capsys.readouterr()
    assert long_message not in captured.out
    with open(message_log, "r") as f:
        assert long_message not in f.read()

    # Re-enables console and tests long message again
    console.enable()
    long_message = "This is a very long message " * 20
    result = console.echo(long_message, LogLevel.INFO, terminal=True, log=True)
    assert result is True
    captured = capsys.readouterr()

    if backend == LogBackends.LOGURU:
        # Removes ANSI color codes, timestamps, and log levels
        cleaned_output = re.sub(r"\x1b\[[0-9;]*m", "", captured.out)
        cleaned_output = re.sub(r"\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{3} \| \w+\s+\| ", "", cleaned_output)
        # Remove extra spaces and newlines
        cleaned_output = " ".join(cleaned_output.split())
    else:  # CLICK
        # Removes extra spaces and newlines
        cleaned_output = " ".join(captured.out.split())

    # Removes extra spaces from long_message for comparison
    cleaned_long_message = " ".join(long_message.split())

    assert cleaned_long_message in cleaned_output

    # Checks log file
    with open(message_log, "r") as f:
        log_content = f.read()
        assert cleaned_long_message in " ".join(log_content.split())


@pytest.mark.parametrize("backend", [LogBackends.LOGURU, LogBackends.CLICK])
def test_console_echo_errors(backend, tmp_path, capsys):
    """Verifies the echo() error-handling functionality."""
    console = Console(
        logger_backend=backend,
        debug_log_path=tmp_path / "debug.log",
        message_log_path=tmp_path / "message.log",
        error_log_path=tmp_path / "error.log",
    )
    console.enable()
    console.add_handles()

    # Tests invalid message type
    with pytest.raises(ValidationError):
        # noinspection PyTypeChecker
        console.echo(message=123, level=LogLevel.INFO)

    # Tests invalid level type
    with pytest.raises(ValidationError):
        # noinspection PyTypeChecker
        console.echo(message="Test message", level="INFO")

    # Tests invalid, but bool-convertible terminal flag type
    # noinspection PyTypeChecker
    result = console.echo(message="Test message", level=LogLevel.INFO, terminal="True")
    assert result is True  # Assuming "True" is converted to True

    # Tests invalid, but bool-convertible log flag type
    # noinspection PyTypeChecker
    result = console.echo(message="Test message", level=LogLevel.INFO, log="True")
    assert result is True  # Assuming "True" is converted to True

    # Tests non-boolean invalid flag handling
    with pytest.raises(ValidationError):
        # noinspection PyTypeChecker
        console.echo(message="Test message", level=LogLevel.INFO, log="123")

    # Tests error when using Loguru backend without handles
    if backend == LogBackends.LOGURU:
        console._logger.remove()  # Remove all handlers
        message = (
            f"Unable to echo the requested message: {'Test message'}. The Console class is configured to use the "
            f"loguru backend, but it does not have any handles. Call add_handles() method to add "
            f"handles or disable() to disable Console operation."
        )
        with pytest.raises(RuntimeError, match=error_format(message)):
            console.echo("Test message", LogLevel.INFO)

        console.add_handles()  # Re-adds handles for the tests below work

    # Tests behavior when console is disabled
    console.disable()
    result = console.echo("Test message", LogLevel.INFO)
    assert result is False

    # Re-enables and test valid call (should not raise any exception)
    console.enable()
    try:
        result = console.echo("Test message", LogLevel.INFO)
        assert result is True
    except Exception as e:
        pytest.fail(f"Valid echo call raised an unexpected exception: {e}")

    # Capture and check the output
    captured = capsys.readouterr()
    if backend == LogBackends.LOGURU:
        assert "Test message" in captured.out
    else:  # CLICK
        assert "Test message" in captured.out

    # Clear captured output
    capsys.readouterr()


@pytest.mark.parametrize("backend", [LogBackends.LOGURU, LogBackends.CLICK])
def test_console_error(backend, tmp_path, capsys):
    """Verifies that the error() console method functions as expected for all backends."""
    console = Console(logger_backend=backend, error_log_path=tmp_path / "error.log")
    console.enable()
    console.add_handles(error_terminal=True, error_file=True)

    # Tests successful error raising and logging
    with pytest.raises(RuntimeError, match="Test error"):
        console.error("Test error", RuntimeError, callback=None, terminal=True, log=True, reraise=True)

    captured = capsys.readouterr()
    assert "Test error" in captured.err
    assert os.path.exists(tmp_path / "error.log")
    with open(tmp_path / "error.log", "r") as f:
        assert "Test error" in f.read()

    # Tests error without reraise
    console.error(message="No reraise error", error=ValueError, callback=None, terminal=True, log=True, reraise=False)
    captured = capsys.readouterr()
    assert "No reraise error" in captured.err
    with open(tmp_path / "error.log", "r") as f:
        assert "No reraise error" in f.read()

    # Tests custom exception
    class CustomError(Exception):
        pass

    with pytest.raises(CustomError):
        console.error(message="Custom error", error=CustomError, callback=None, terminal=True, log=True, reraise=True)

    # Capture and check the output
    captured = capsys.readouterr()
    assert "Custom error" in captured.err

    # Tests callback (only for Loguru backend)
    if backend == LogBackends.LOGURU:
        callback_called = False

        def callback_func(_error):
            nonlocal callback_called
            callback_called = True
            print("Callback executed", file=sys.stderr)

        # noinspection PyTypeChecker
        console.error(
            "Callback error", error=ValueError, callback=callback_func, terminal=True, log=True, reraise=False
        )
        captured = capsys.readouterr()
        assert "Callback error" in captured.err
        assert "Callback executed" in captured.err
        assert callback_called

        # Also tests default callback performance (that it correctly aborts the runtime)
        with pytest.raises(SystemExit):
            console.error("Callback error", error=ValueError, terminal=False, log=False, reraise=False)

    # Checks that all errors were logged
    with open(tmp_path / "error.log", "r") as f:
        log_content = f.read()
        assert "Test error" in log_content
        assert "No reraise error" in log_content
        assert "Custom error" in log_content
        if backend == LogBackends.LOGURU:
            assert "Callback error" in log_content


@pytest.mark.parametrize("backend", [LogBackends.LOGURU, LogBackends.CLICK])
def test_console_error_output_options(backend, tmp_path, capsys):
    """Verifies that different error logging destinations work as expected."""
    error_log = tmp_path / "error.log"
    console = Console(logger_backend=backend, error_log_path=error_log)
    console.enable()
    console.add_handles(error_terminal=True, error_file=True)

    # Tests terminal only
    console.error("Terminal only", RuntimeError, callback=None, terminal=True, log=False, reraise=False)
    captured = capsys.readouterr()
    assert "Terminal only" in captured.err

    # Loguru automatically creates an empty log file, click backend does not create the file
    if backend == LogBackends.LOGURU:
        assert error_log.read_text() == ""
    else:
        assert not os.path.exists(error_log)

    # Tests log only
    console.error("Log only", RuntimeError, callback=None, terminal=False, log=True, reraise=False)
    captured = capsys.readouterr()
    assert captured.err == ""
    with open(tmp_path / "error.log", "r") as f:
        assert "Log only" in f.read()

    # Tests both terminal and log output disabled
    console.error("Neither", RuntimeError, callback=None, terminal=False, log=False, reraise=False)
    captured = capsys.readouterr()
    assert captured.err == ""
    with open(tmp_path / "error.log", "r") as f:
        assert "Neither" not in f.read()


@pytest.mark.parametrize("backend", [LogBackends.LOGURU, LogBackends.CLICK])
def test_console_error_handling(backend, tmp_path):
    """Verifies error-handling functionality for the error() method."""
    console = Console(logger_backend=backend, error_log_path=tmp_path / "error.log")
    console.enable()
    console.add_handles(error_terminal=True, error_file=True)

    # Tests disabled console. Disabled console is still expected to raise the error itself, but should not do any
    # additional processing.
    console.disable()
    with pytest.raises(RuntimeError, match=error_format("Disabled error")):
        console.error("Disabled error", RuntimeError, terminal=True, log=True)
    if backend == LogBackends.LOGURU:
        assert Path(tmp_path / "error.log").read_text() == ""
    else:
        assert not os.path.exists(tmp_path / "error.log")

    # Re-enables console for further tests
    console.enable()

    # Tests error handling
    with pytest.raises(ValidationError):
        # noinspection PyTypeChecker
        console.error(123, RuntimeError)  # Invalid message type

    with pytest.raises(ValidationError):
        # noinspection PyTypeChecker
        console.error("Test", "Not a callable")  # Invalid error type

    # Tests Loguru-specific error (no handles)
    if backend == LogBackends.LOGURU:
        console._logger.remove()
        message = (
            f"Unable to properly log the requested error ({RuntimeError}) with message {'No handles error'}. The "
            f"Console class is configured to use the loguru backend, but it does not have any handles. "
            f"Call add_handles() method to add handles or disable() to disable Console operation."
        )
        with pytest.raises(RuntimeError, match=error_format(message)):
            console.error("No handles error", RuntimeError)


def test_ensure_directory_exists() -> None:
    """Verifies that _ensure_directory_exists() method of Console class functions as expected"""
    with tempfile.TemporaryDirectory() as temp_dir:
        # Test with a directory path
        dir_path = Path(temp_dir) / "test_dir"
        Console._ensure_directory_exists(dir_path)
        assert dir_path.exists() and dir_path.is_dir()

        # Test with a file path
        file_path = Path(temp_dir) / "nested" / "dir" / "test_file.txt"
        Console._ensure_directory_exists(file_path)
        assert file_path.parent.exists() and file_path.parent.is_dir()
        assert not file_path.exists()  # The file itself should not be created

        # Test with an existing directory
        existing_dir = Path(temp_dir) / "existing_dir"
        existing_dir.mkdir()
        Console._ensure_directory_exists(existing_dir)
        assert existing_dir.exists() and existing_dir.is_dir()

        # Test with a path that includes a file in an existing directory
        existing_file_path = Path(temp_dir) / "test_file2.txt"
        Console._ensure_directory_exists(existing_file_path)
        assert existing_file_path.parent.exists() and existing_file_path.parent.is_dir()
        assert not existing_file_path.exists()  # The file itself should not be created

        # Test with a deeply nested path
        deep_path = Path(temp_dir) / "very" / "deep" / "nested" / "directory" / "structure"
        Console._ensure_directory_exists(deep_path)
        assert deep_path.exists() and deep_path.is_dir()


@pytest.fixture
def console() -> Console:
    """This fixture returns Console class instance initialized with default settings.

    It is used to verify basic Console properties and methods.
    """
    return Console()


def test_debug_log_path(console, tmp_path):
    """Verifies the functionality of debug_log_path getter and setter methods."""
    # Tests getter when the path is not set
    assert console.get_debug_log_path is None

    # Tests setter and getter
    debug_path = tmp_path / "debug.log"
    console.set_debug_log_path(debug_path)
    assert console.get_debug_log_path == debug_path


def test_message_log_path(console, tmp_path):
    """Verifies the functionality of message_log_path getter and setter methods."""
    # Tests getter when the path is not set
    assert console.get_message_log_path is None

    # Tests setter and getter
    message_path = tmp_path / "message.log"
    console.set_message_log_path(message_path)
    assert console.get_message_log_path == message_path


def test_error_log_path(console, tmp_path):
    """Verifies the functionality of error_log_path getter and setter methods."""
    # Tests getter when the path is not set
    assert console.get_error_log_path is None

    # Tests setter and getter
    error_path = tmp_path / "error.log"
    console.set_error_log_path(error_path)
    assert console.get_error_log_path == error_path


def test_invalid_path_error_handling(console):
    """Verifies that log path setter methods correctly raise ValueError when provided with an invalid path."""
    invalid_paths = [
        Path("invalid.zippp"),  # Invalid extension
        Path("invalid"),  # No extension
    ]

    for invalid_path in invalid_paths:
        with pytest.raises(ValueError):
            console.set_debug_log_path(invalid_path)

        with pytest.raises(ValueError):
            console.set_message_log_path(invalid_path)

        with pytest.raises(ValueError):
            console.set_error_log_path(invalid_path)



"""This module contains the Console class that provides message and error terminal-printing and file-logging
functionality.

Additionally, it contains enumerations and standalone functions that are used to support Console class runtimes.
"""

import sys
from enum import Enum
from types import NoneType
from typing import Any, Literal, Optional
from pathlib import Path
import textwrap
from dataclasses import fields, dataclass
from collections.abc import Callable

import click
from loguru import logger
from pydantic import validate_call

# noinspection PyProtectedMember
from loguru._logger import Logger


def default_callback(__error: str | int | None = None) -> Any:
    """Calls sys.exit() with a minimal explanation code.

    This is a wrapper over sys.exit() that can be used as the input to 'onerror' argument of loguru catch() method.
    The main advantage of using this callback over the plain sys.exit is that it avoids reprinting the exception
    message, reducing the output clutter.
    """
    sys.exit("Runtime aborted.")


class LogLevel(Enum):
    """Maps valid literal arguments that can be passed to some Console class methods to programmatically callable
    variables.

    Use this enumeration instead of 'hardcoding' logging levels where possible to automatically adjust to future API
    changes of this library.

    Log level determines the 'severity' of the logged messages. In turn, this is used to conditionally filter incoming
    messages, depending on the configuration of the Console class loguru backend. For example, the end-user can disable
    the handles for 'DEBUG' level messages and suppress any message at or below DEBUG level.
    """

    DEBUG: str = "debug"
    """
    Messages that are not shown by default and need to be purposefully enabled. These messages can be left
    in source code during early project stages to speed-up debugging, but, ideally, should be removed for mature
    projects.
    """
    INFO: str = "info"
    """
    General information messages, such as reporting the progress of a runtime.
    """
    SUCCESS: str = "success"
    """
    Runtime-ending messages specifically informing that a runtime ran successfully.
    """
    WARNING: str = "warning"
    """
    Non-runtime-breaking, but potentially problematic messages, such as deprecation warnings.
    """
    ERROR: str = "error"
    """
    Typically used when dealing with exceptions. Report runtime-breaking errors and are typically augmented with 
    traceback information.
    """
    CRITICAL: str = "critical"
    """
    Same as ERROR level, but more important. Critical errors are displayed using an augmented style that draws 
    end-user attention. Generally, this level is not used in most production runtimes.
    """


class LogBackends(Enum):
    """Maps valid backend options that can be used to instantiate the Console class to programmatically addressable
    variables.

    Use this enumeration to specify the backend used by the Console class to display and save logged messages to files.

    The backend determines the message and error processing engine used by the Console class. For most projects, it
    is highly advised to use the default loguru backend as it provides a more robust feature set, compared to
    'click' backend.
    """

    LOGURU: str = "loguru"
    """
    Loguru is the default backend for handling message and error printing and logging as it provides a robust set of 
    features and a high degree of customization. The Console class was primarily written to work with this backend.
    """
    CLICK: str = "click"
    """
    The backup backend, which also allows printing and logging messages and errors, but is not as robust as loguru.
    """


@dataclass
class LogExtensions:
    """Maps valid file-extension options that can be used by log file paths provided to the Console class to
    programmatically addressable variables.

    Use this class to add valid extensions to the log-file paths used as input arguments when initializing new
    Console class instances or augmenting existing Console class instances via setter methods.

    File extensions are used to determine the log file format. Extensions exposed through this class already contain
    the '.' prefix and should be appended to plain file names. For example, to add .log extension, you can use:
    f"file_name{LogExtensions.LOG}"
    """

    LOG: str = ".log"
    """
    Log file extensions should be the default for human-readable log files according to the general convention. These 
    files will behave exactly like .txt files, but their extension will further emphasize that they are log dumps.
    """
    TXT: str = ".txt"
    """
    While generally discouraged, the default text extension can also be used for log dump files. These files will behave
    like any other text file.
    """
    JSON: str = ".json"
    """
    A special log file extension that is generally preferred for logs that are intended to be parsed from software. 
    Unlike other supported extensions, .json files are not directly human-readable, but provide better support for 
    programmatically parsing the logged data.
    """

    @classmethod
    def values(cls) -> tuple[str, ...]:
        """Returns the valid extension options packaged into a tuple.

        The returned tuple is used by the Console class to validate incoming log paths.
        """
        return tuple(getattr(cls, field.name) for field in fields(cls))


class Console:
    """After initial configuration, provides methods for terminal-printing and file-logging messages and errors.

    This class wraps multiple message-processing (logging and / or printing) backends and provides an API that allows
    configuring and using the wrapped backend in a consistent fashion across many projects. Overall, it is designed to
    largely behave like the standard 'print()' and 'raise' methods offered by the default Python library.

    Notes:
        Since this class is explicitly designed to be shared by multiple projects that may also be mutually-dependent,
        it defaults to a disabled state. When Console is initialized, calling its echo() (analogous to print()) method
        will not produce any output and calling error() (analogous to raise) method will behave like a standard
        'raise' method. To enable the full class functionality, the Console has to be configured (via add_handles() and
        enabled (via enable()) methods.

        Do not configure or enable the Console class from libraries that may be imported by other projects! To work
        properly, the Console has to be enabled at the highest level of the call hierarchy: from the main runtime
        script. Leave console configuration and enabling to the end-user.

    Args:
        logger_backend: Specifies the backend used to process message and error terminal-printing and file-logging.
            Valid backend options are available through LogBackends enumeration and are currently limited to
            LOGURU and CLICK.
        line_width: The maximum length, in characters, for a single line of displayed text. This is used to limit the
            width of the text block as it is displayed in the terminal and written to log files.
        debug_log_path: The path to the file used to log debug messages (messages at or below DEBUG level). If not
            provided (set to None), logging debug messages will be disabled.
        message_log_path: The path to the file used to log non-error messages (INFO through WARNING levels). If not
            provided (set to None), logging non-debug messages will be disabled.
        error_log_path: The path to the file used to log errors (messages at or above ERROR level). If not provided
            (set to None), logging errors will be disabled.
        break_long_words: Determines whether to break long words when formatting the text block to fit the width
            requirement.
        break_on_hyphens: Determines whether to break sentences on hyphens when formatting the text block to fit the
            width requirement.
        use_color: Determines whether to colorize the terminal output. This primarily applies to loguru backend.

    Attributes:
        _line_width: Stores the maximum allowed text block line width, in characters.
        _break_long_words: Determines whether to break text on long words.
        _break_on_hyphens: Determines whether to break text on hyphens.
        _use_color: Determines whether to colorize terminal-printed and file-logged text.
        _valid_extensions: Stores valid log-file extensions. This is used to verify input log file paths, as valid paths
            are expected to end with one of the supported extensions.
        _debug_log_path: Stores the path to the debug log file.
        _message_log_path: Stores the path to the message log file.
        _error_log_path: Stores the path to the error log file.
        _backend: Stores the backend option used to provide the terminal-printing and file-logging functionality.
        _logger: When logging backend is set to LOGURU, stores the instance of the loguru 'Logger' class. Otherwise, it
            is set to None.
        _is_enabled: Tracks whether logging through this class instance is enabled. When this tracker is False, echo()
            and print() methods will have limited or no functionality.

    Raises:
        ValueError: If any of the provided log file paths is not valid.
        ValidationError: If any of the input arguments are not of a valid type.
    """

    @validate_call()
    def __init__(
        self,
        logger_backend: Literal[LogBackends.LOGURU, LogBackends.CLICK] = LogBackends.LOGURU,
        debug_log_path: Optional[Path | str] = None,
        message_log_path: Optional[Path | str] = None,
        error_log_path: Optional[Path | str] = None,
        line_width: int = 120,
        break_long_words: bool = False,
        break_on_hyphens: bool = False,
        use_color: bool = True,
    ) -> None:
        # Message formating parameters.
        if line_width <= 0:
            message = (
                f"Invalid 'line_width' argument encountered when instantiating Console class instance. "
                f"Expected a value greater than 0, but encountered {line_width}."
            )
            raise ValueError(
                textwrap.fill(
                    text=message, width=120, break_on_hyphens=break_on_hyphens, break_long_words=break_long_words
                )
            )
        self._line_width: int = line_width
        self._break_long_words: bool = break_long_words
        self._break_on_hyphens: bool = break_on_hyphens
        self._use_color: bool = use_color

        self._valid_extensions: tuple[str, ...] = LogExtensions.values()

        # Verifies that the input paths to log files, if any, use valid file extensions and are otherwise well-formed.
        # Stores currently supported log file extensions
        if not isinstance(debug_log_path, NoneType):
            debug_log_path = Path(debug_log_path)
            if debug_log_path.suffix not in self._valid_extensions:
                message = (
                    f"Invalid 'debug_log_path' argument encountered when instantiating Console class instance. "
                    f"Expected a path ending in a file name with one of the supported extensions:"
                    f"{', '.join(self._valid_extensions)}, but encountered {debug_log_path}."
                )
                raise ValueError(
                    textwrap.fill(
                        text=message,
                        width=self._line_width,
                        break_on_hyphens=self._break_on_hyphens,
                        break_long_words=self._break_long_words,
                    )
                )
            # If the path is valid, verifies the directory portion of the path exists and, if not, creates it.
            self._ensure_directory_exists(debug_log_path)

        if not isinstance(message_log_path, NoneType):
            message_log_path = Path(message_log_path)
            if message_log_path.suffix not in self._valid_extensions:
                message = (
                    f"Invalid 'message_log_path' argument encountered when instantiating Console class instance. "
                    f"Expected a path ending in a file name with one of the supported extensions:"
                    f"{', '.join(self._valid_extensions)}, but encountered {message_log_path}."
                )
                raise ValueError(
                    textwrap.fill(
                        text=message,
                        width=self._line_width,
                        break_on_hyphens=self._break_on_hyphens,
                        break_long_words=self._break_long_words,
                    )
                )
            self._ensure_directory_exists(message_log_path)

        if not isinstance(error_log_path, NoneType):
            error_log_path = Path(error_log_path)
            if error_log_path.suffix not in self._valid_extensions:
                message = (
                    f"Invalid 'error_log_path' argument encountered when instantiating Console class instance. "
                    f"Expected a path ending in a file name with one of the supported extensions:"
                    f"{', '.join(self._valid_extensions)}, but encountered {error_log_path}."
                )
                raise ValueError(
                    textwrap.fill(
                        text=message,
                        width=self._line_width,
                        break_on_hyphens=self._break_on_hyphens,
                        break_long_words=self._break_long_words,
                    )
                )

            self._ensure_directory_exists(error_log_path)

        self._debug_log_path: Optional[Path] = debug_log_path
        self._message_log_path: Optional[Path] = message_log_path
        self._error_log_path: Optional[Path] = error_log_path

        # Internal trackers
        self._backend = logger_backend
        if logger_backend == LogBackends.LOGURU:
            self._logger: Optional[Logger] = logger  # type: ignore
        else:
            self._logger = None
        self._is_enabled: bool = False

    @validate_call()
    def add_handles(
        self,
        *,
        remove_existing_handles: bool = True,
        debug_terminal: bool = False,
        debug_file: bool = False,
        message_terminal: bool = True,
        message_file: bool = False,
        error_terminal: bool = True,
        error_file: bool = False,
        enqueue: bool = False,
    ) -> None:
        """(Re)configures the local loguru 'logger' instance to use requested handles after optionally removing all
        existing handles.

        This method is only used when Console uses 'loguru' backend. It has no effect for other backends.

        The handles control which messages (levels) can be processed and where they are sent (terminal, file, or both).
        This method adds two separate handles to save messages to files and print them to the terminal
        for each of the 3 supported level categories: at or below DEBUG, INFO through WARNING, and at or above ERROR.
        Overall, this means the method can add up to 6 handles.

        This method only needs to be called once and only from the highest level of the call hierarchy, such as the
        main runtime script or module. Do not call this method from libraries designed to be used in other projects to
        avoid interfering with upstream processes instantiating their own handles.

        Notes:
            The method can be flexibly configured to only add a subset of all supported handles. For example,
            by default, it does not add debug handles, making it impossible to terminal-print or file-log debug
            messages. It can also be configured to not remove existing handles (default behavior) if necessary. See
            argument docstrings below for more information.

            During runtime, handles determine what happens to the message passed via the appropriate 'log' call. Loguru
            shares the set of handles across all 'logger' instances, which means this method should be used with
            caution, as it can interfere with any other handles, including the default ones.

        Args:
            remove_existing_handles: Determines whether to remove all existing handles before adding new loguru handles.
                Since loguru comes with 'default' handles enabled, this is almost always the recommended option.
            debug_terminal: Determines whether to add the handle that prints messages at or below DEBUG level to
                terminal.
            debug_file: Determines whether to add the handle that writes messages at or below DEBUG level to debug-log
                file.
            message_terminal: Same as debug_terminal, but for messages at INFO through WARNING levels.
            message_file: Same as debug_file, but for messages at INFO through WARNING levels.
            error_terminal: Same as debug_terminal, but for messages at or above ERROR level.
            error_file: Same as debug_file, but for messages at or above ERROR level.
            enqueue: Determines if messages are processed synchronously or asynchronously. Generally, this option is
                only suggested for multiprocessing runtimes that handle messages from multiple processes, as queueing
                messages prevents common multiprocessing / multithreading issues such as race conditions.

        Raises:
            ValidationError: If any of the input arguments are not of a valid type.
        """
        # Returns immediately for non-loguru Consoles.
        if self._backend != LogBackends.LOGURU or isinstance(self._logger, NoneType):
            return

        # If necessary, removes existing handles.
        if remove_existing_handles:
            self._logger.remove()

        # Adds ataraxis_shell and ataraxis_log flags to the 'extras' dictionary passed together with each logged
        # message. These flags are used to dynamically control whether every processed message should be sent to
        # the terminal, written to the file or both.
        self._logger = self._logger.bind(ataraxis_shell=False, ataraxis_log=False)

        # Mostly here to prevent mypy being annoying, as the error is not really possible
        if isinstance(self._logger, NoneType):
            message = (
                "Unable to bind the logger to use the required extra variables. Generally, this error should not be "
                "possible"
            )
            raise RuntimeError(
                textwrap.fill(
                    text=message,
                    width=self._line_width,
                    break_on_hyphens=self._break_on_hyphens,
                    break_long_words=self._break_long_words,
                )
            )

        # Debug terminal-printing handle. Filters and works for any message with the log-level at or below DEBUG.
        # Includes 'diagnose' information, which provides additional information about the objects involved in
        # generating the message. Also uses 'ataraxis_shell' extra tag to determine if any message should be processed
        # or not.
        if debug_terminal:
            # noinspection LongLine
            self._logger.add(
                sys.stdout,
                format="<magenta>{time:YYYY-MM-DD HH:mm:ss.SSS}</magenta> | <level>{level: <8}</level> | <level>{message}</level>",
                filter=lambda record: record["extra"]["ataraxis_terminal"] is True
                and record["level"].no <= logger.level("DEBUG").no,
                colorize=True,
                backtrace=False,
                diagnose=True,
                enqueue=enqueue,
            )

        # Message terminal-printing handle. Functions as a prettier, time-stamped print. Does not include any additional
        # information and only prints messages with level above DEBUG and up to WARNING (inclusive).
        if message_terminal:
            # noinspection LongLine
            self._logger.add(
                sys.stdout,
                format="<magenta>{time:YYYY-MM-DD HH:mm:ss.SSS}</magenta> | <level>{level: <8}</level> | <level>{message}</level>",
                filter=lambda record: record["extra"]["ataraxis_terminal"] is True
                and logger.level("WARNING").no >= record["level"].no > logger.level("DEBUG").no,
                colorize=True,
                backtrace=False,
                diagnose=False,
                enqueue=enqueue,
            )

        # Error terminal-printing-handle. Does not include additional diagnostic information, but includes the whole
        # backtrace of the error message. It works similarly to default python error traces, but without mandatory
        # runtime termination. Works for ERROR+ level messages. Unlike other two handles, writes to
        # stderr, rather than stdout.
        if error_terminal:
            # noinspection LongLine
            self._logger.add(
                sys.stderr,
                format="<magenta>{time:YYYY-MM-DD HH:mm:ss.SSS}</magenta> | <level>{level: <8}</level> | <level>{message}</level>",
                filter=lambda record: record["extra"]["ataraxis_terminal"] is True
                and record["level"].no > logger.level("WARNING").no,
                colorize=True,
                backtrace=True,
                diagnose=False,
                enqueue=enqueue,
            )

        # Debug file-writing handle. The only difference from the terminal handle is that it writes to a file, rather
        # than the stdout handle and that ut uses ataraxis_log tag instead of the ataraxis_shell. Debug files are
        # automatically removed after 2 days and are not compressed as they are considered temporary.
        if not isinstance(self._debug_log_path, NoneType) and debug_file:
            self._logger.add(
                self._debug_log_path,
                filter=lambda record: record["extra"]["ataraxis_log"] is True
                and record["level"].no <= logger.level("DEBUG").no,
                colorize=False,
                retention="2 days",
                rotation="500 MB",
                enqueue=enqueue,
            )

        # Message file-writing handle. Functions similarly to terminal-printing handle, but prints to a file that does
        # not have a rotation window and is retained forever.
        if not isinstance(self._message_log_path, NoneType) and message_file:
            self._logger.add(
                self._message_log_path,
                filter=lambda record: record["extra"]["ataraxis_log"] is True
                and logger.level("WARNING").no >= record["level"].no > logger.level("DEBUG").no,
                colorize=False,
                enqueue=enqueue,
            )

        # Error file-writing handle. Error files are rotated once they reach 100 MB and only retained for 5 days.
        # In addition to the full traceback, the logs include diagnostic information that provides data about objects
        # along the execution stack that led to an error to allow in-depth analysis of the problem.
        if not isinstance(self._error_log_path, NoneType) and error_file:
            self._logger.add(
                self._error_log_path,
                filter=lambda record: record["extra"]["ataraxis_log"] is True
                and record["level"].no >= logger.level("ERROR").no,
                colorize=False,
                backtrace=True,
                diagnose=True,
                rotation="100 MB",
                retention="5 days",
                enqueue=enqueue,
            )

    def enable(self) -> None:
        """Enables processing messages and errors with this Console class."""
        self._is_enabled = True

    def disable(self) -> None:
        """Disables processing messages and errors with this Console class.

        Notes:
            Even when console is disabled, the error() method will still raise exceptions, but will not log them to
            files or provide detailed traceback information.
        """
        self._is_enabled = False

    @property
    def get_debug_log_path(self) -> Path | None:
        """Returns the path to the log file used to save messages at or below DEBUG level or None if the path was not
        set."""
        return self._debug_log_path

    @validate_call()
    def set_debug_log_path(self, path: Path) -> None:
        """Sets the path to the log file used to save messages at or below DEBUG level.

        Notes:
            Remember to call add_handles() method to reconfigure the handles after providing the new path when using
            loguru backend.

        Raises:
            ValueError: If the provided path does not end with one of the supported file-extensions.
        """
        # Verifies tha the path points ot a valid file
        if path.suffix not in self._valid_extensions:
            message = (
                f"Invalid 'path' argument encountered when setting Console debug_log_path. "
                f"Expected a path ending in a file name with one of the supported extensions:"
                f"{', '.join(self._valid_extensions)}, but encountered {path}."
            )
            raise ValueError(
                textwrap.fill(
                    text=message,
                    width=self._line_width,
                    break_on_hyphens=self._break_on_hyphens,
                    break_long_words=self._break_long_words,
                )
            )

        # Ensures that the directory included in the path exists and overwrites the local debug log path
        self._ensure_directory_exists(path)
        self._debug_log_path = path

    @property
    def get_message_log_path(self) -> Path | None:
        """Returns the path to the log file used to save messages at INFO through WARNING levels or None if the path
        was not set."""
        return self._message_log_path

    @validate_call()
    def set_message_log_path(self, path: Path) -> None:
        """Sets the path to the log file used to save messages at INFO through WARNING levels.

        Notes:
            Remember to call add_handles() method to reconfigure the handles after providing the new path when using
            loguru backend.

        Raises:
            ValueError: If the provided path does not end with one of the supported file-extensions.
        """
        if path.suffix not in self._valid_extensions:
            message = (
                f"Invalid 'path' argument encountered when setting Console message_log_path. "
                f"Expected a path ending in a file name with one of the supported extensions:"
                f"{', '.join(self._valid_extensions)}, but encountered {path}."
            )
            raise ValueError(
                textwrap.fill(
                    text=message,
                    width=self._line_width,
                    break_on_hyphens=self._break_on_hyphens,
                    break_long_words=self._break_long_words,
                )
            )

        # Ensures that the directory included in the path exists and overwrites the local message log path
        self._ensure_directory_exists(path)
        self._message_log_path = path

    @property
    def get_error_log_path(self) -> Path | None:
        """Returns the path to the log file used to save messages at or above ERROR level or None if the path was not
        set."""
        return self._error_log_path

    @validate_call()
    def set_error_log_path(self, path: Path) -> None:
        """Sets the path to the log file used to save messages at or above ERROR level.

        Notes:
            Remember to call add_handles() method to reconfigure the handles after providing the new path when using
            loguru backend.

        Raises:
            ValueError: If the provided path does not end with one of the supported file-extensions.
        """
        if path.suffix not in self._valid_extensions:
            message = (
                f"Invalid 'path' argument encountered when setting Console error_log_path. "
                f"Expected a path ending in a file name with one of the supported extensions:"
                f"{', '.join(self._valid_extensions)}, but encountered {path}."
            )
            raise ValueError(
                textwrap.fill(
                    text=message,
                    width=self._line_width,
                    break_on_hyphens=self._break_on_hyphens,
                    break_long_words=self._break_long_words,
                )
            )

        # Ensures that the directory included in the path exists and overwrites the local error log path
        self._ensure_directory_exists(path)
        self._error_log_path = path

    @property
    def has_handles(self) -> bool:
        """Returns True if the class uses loguru backend and the backend has configured handles.

        If the class does not use loguru backend or if the class uses loguru and does not have handles, returns
        False."""
        if self._backend == LogBackends.LOGURU and not isinstance(self._logger, NoneType):
            # noinspection PyProtectedMember
            return len(self._logger._core.handlers) > 0
        else:
            return False

    @property
    def is_enabled(self) -> bool:
        """Returns True if logging with this Console class instance is enabled."""
        return self._is_enabled

    @staticmethod
    def _ensure_directory_exists(path: Path) -> None:
        """Determines if the directory portion of the input path exists and, if not, creates it.

        When the input path ends with an .extension (indicating a file path), the file portion is ignored and
        only the directory path is evaluated.

        Args:
            path: The path to be processed. Can be a file or a directory path.
        """
        # If the path is a file (because it has a .extension suffix), ensures the parent directory of the file, if any,
        # exists.
        if path.suffix != "":
            path.parent.mkdir(parents=True, exist_ok=True)
        else:
            # If the path is a directory path, ensures the directory exists.
            path.mkdir(parents=True, exist_ok=True)

    @validate_call()
    def format_message(self, message: str, *, loguru: bool = False) -> str:
        """Formats the input message string according to the class configuration parameters.

        This method is generally intended to be used internally as part of the echo() or error() method runtimes.
        However, it can also be accessed and used externally to maintain consistent text formatting across the
        application.

        Args:
            message: The text string to format according to class configuration parameters.
            loguru: A flag that determines if the message is intended to be subsequently processed via loguru backend or
                another method or backend (e.g.: Exception class or CLICK backend).

        Returns:
            Formatted text message (augmented with newline and other service characters as necessary).

        Raises:
            ValidationError: If any of the arguments are not of a valid type.
        """

        # For loguru-processed messages, uses a custom formatting that accounts for the prepended header. The header
        # is assumed to be matching the standard defined in add_handles() method, which statically reserves 37
        # characters of the first line.
        if loguru:
            # Calculates indent and dedent parameters for the lines
            first_line_width: int = self._line_width - 37  # Makes the first line shorter
            subsequent_indent: str = " " * 37
            lines: list[str] = []

            # Handles the first line by wrapping it to fit into the required width given the additional loguru header.
            first_line: str = message[:first_line_width]  # Subtracts loguru header
            if len(message) > first_line_width:  # Determines the wrapping point
                # Finds the last space in the first line to avoid breaking words
                last_space: int = first_line.rfind(" ")
                if last_space != -1:  # Wraps the line
                    first_line = first_line[:last_space]

            lines.append(first_line)

            # Wraps the rest of the message by statically calling textwrap.fill on it with precalculated indent to align
            # the text to the first line.
            rest_of_message: str = message[len(first_line) :].strip()
            if rest_of_message:
                subsequent_lines = textwrap.fill(
                    rest_of_message,
                    width=self._line_width,
                    initial_indent=subsequent_indent,
                    subsequent_indent=subsequent_indent,
                    break_long_words=self._break_long_words,
                    break_on_hyphens=self._break_on_hyphens,
                )
                lines.extend(subsequent_lines.splitlines())

            return "\n".join(lines)

        # For non-loguru-processed messages, simply wraps the message via textwrap.
        else:
            return textwrap.fill(
                text=message,
                width=self._line_width,
                break_long_words=self._break_long_words,
                break_on_hyphens=self._break_on_hyphens,
            )

    @validate_call()
    def echo(self, message: str, level: LogLevel = LogLevel.INFO, *, terminal: bool = True, log: bool = True) -> bool:
        """Formats the input message according to the class configuration and outputs it to the terminal, file, or both.

        In a way, this can be seen as a better 'print'. Specifically, in addition to printing the text to the terminal,
        this method supports colored output and can simultaneously print the message to the terminal and write it to a
        log file.

        Args:
            message: The message to be processed.
            level: The severity level of the message. This method supports all levels available through the LogLevel
                enumeration, but is primarily intended to be used for DEBUG, INFO, SUCCESS, and WARNING messages.
                Errors should be raised through the error() method when possible.
            terminal: The flag that determines whether the message should be printed to the terminal using the class
                logging backend. For loguru backend, this acts on top of the handle configuration. If there are no
                handles to print the message to the terminal, the value of this flag is ignored.
            log: The flag that determines whether the message should be written to a log file using the class logging
                backend. For loguru backend, this acts on top of the handle configuration. If there are no
                handles to save the message to a log file, the value of this flag is ignored. Note, the handle
                configuration is in turn dependent on whether valid log file path(s) were provided to the class before
                calling add_handles() method.

        Returns:
            True if the message has been processed and False if the message cannot be printed because the Console is
            disabled.

        Raises:
            ValidationError: If any of the input arguments are not of a valid type.
            RuntimeError: If the method is called while using loguru backend without any active logger handles.
        """

        # If the Console is disabled, returns False
        if not self.is_enabled:
            return False

        # Loguru backend
        if self._backend == LogBackends.LOGURU and not isinstance(self._logger, NoneType):
            if not self.has_handles:
                message = (
                    f"Unable to echo the requested message: {message}. The Console class is configured to use the "
                    f"loguru backend, but it does not have any handles. Call add_handles() method to add "
                    f"handles or disable() to disable Console operation."
                )
                raise RuntimeError(
                    textwrap.fill(
                        text=message,
                        width=self._line_width,
                        break_on_hyphens=self._break_on_hyphens,
                        break_long_words=self._break_long_words,
                    )
                )

            # Formats the message to work with additional loguru-prepended header.
            formatted_message = self.format_message(message=message, loguru=True)

            # Logs the message using the appropriate channel and file / terminal options.
            self._logger = self._logger.bind(ataraxis_log=log, ataraxis_terminal=terminal)

            # Mostly here to prevent mypy being annoying, as the error is not really possible
            if isinstance(self._logger, NoneType):
                message = (
                    "Unable to bind the logger to use the required extra variables. Generally, this error should "
                    "not be possible."
                )
                raise RuntimeError(
                    textwrap.fill(
                        text=message,
                        width=self._line_width,
                        break_on_hyphens=self._break_on_hyphens,
                        break_long_words=self._break_long_words,
                    )
                )

            # For loguru, the message just needs to be logged. Loguru will use available handles to determine where to
            # route the message.
            if level == LogLevel.DEBUG:
                self._logger.debug(formatted_message)
            elif level == LogLevel.INFO:
                self._logger.info(formatted_message)
            elif level == LogLevel.SUCCESS:
                self._logger.success(formatted_message)
            elif level == LogLevel.WARNING:
                self._logger.warning(formatted_message)
            elif level == LogLevel.ERROR:
                self._logger.error(formatted_message)
            elif level == LogLevel.CRITICAL:
                self._logger.critical(formatted_message)

        elif self._backend == LogBackends.CLICK:
            # Formats the message using non-loguru parameters
            formatted_message = self.format_message(message=message, loguru=False)

            # For click, terminal and log file inputs are processed separately
            if terminal:
                if level == LogLevel.DEBUG:
                    click.secho(message=formatted_message, err=False, color=self._use_color, fg="cyan")
                elif level == LogLevel.INFO:
                    click.secho(message=formatted_message, err=False, color=self._use_color, fg="white")
                elif level == LogLevel.SUCCESS:
                    click.secho(message=formatted_message, err=False, color=self._use_color, fg="green")
                elif level == LogLevel.WARNING:
                    click.secho(message=formatted_message, err=False, color=self._use_color, fg="yellow")
                elif level == LogLevel.ERROR:
                    click.secho(message=formatted_message, err=True, color=self._use_color, fg="red")
                elif level == LogLevel.CRITICAL:
                    click.secho(message=formatted_message, err=True, color=self._use_color, fg="red", bg="white")

            # Does not use colors when writing to log files.
            if log:
                if level == LogLevel.DEBUG and self._debug_log_path:
                    with open(file=str(self._debug_log_path), mode="a") as file:
                        click.echo(file=file, message=formatted_message, color=False)
                elif level == LogLevel.ERROR or level == LogLevel.CRITICAL and self._error_log_path:
                    with open(file=str(self._error_log_path), mode="a") as file:
                        click.echo(file=file, message=formatted_message, color=False)
                elif self._message_log_path:
                    with open(file=str(self._message_log_path), mode="a") as file:
                        click.echo(file=file, message=formatted_message, color=False)

        # Returns true to indicate that the message was processed.
        return True

    @validate_call()
    def error(
        self,
        message: str,
        error: Callable[..., Exception] = RuntimeError,
        callback: Optional[Callable[[], Any]] = default_callback,
        *,
        terminal: bool = True,
        log: bool = True,
        reraise: bool = False,
    ) -> None:
        """Raises the requested error.

        If Console is disabled, this method will format the error message and use the standard Python 'raise' mechanism
        to trigger the requested error. If Console is enabled, the error will be processed in-place according to
        arguments and Console backend configuration.

        Notes:
            When console is enabled, this method can be used to flexibly handle raise errors in-place. For example, it
            can be used to redirect errors to log file, provides enhanced traceback and analysis data (for loguru
            backend only) and can even execute callback functions after logging the error (also for loguru backend only.

        Args:
            message: The error-message to use for the raised error.
            error: The callable Exception class to be raised by the method.
            callback: Optional, only for loguru logging backends. The function to call after catching the raised
                exception. This can be used to terminate or otherwise alter the runtime without relying on the standard
                Python mechanism of retracing the call stack. For example, the default callback terminates the runtime
                in-place, without allowing Python to retrace the call stack that is already traced by loguru.
            terminal: The flag that determines whether the error should be printed to the terminal using the class
                logging backend. For loguru backend, this acts on top of the handle configuration. If there are no
                handles to print the error to the terminal, the value of this flag is ignored.
            log: The flag that determines whether the error should be written to a log file using the class logging
                backend. For loguru backend, this acts on top of the handle configuration. If there are no
                handles to save the error to log file, the value of this flag is ignored. Note, the handle
                configuration is in turn dependent on whether valid log file path(s) were provided to the class before
                calling add_handles() method.
            reraise: The flag that determines whether to reraise the error after it is caught and handled by
                the logging backend. For non-loguru backends, this determines if the error is raised in the first place
                or if the method only logs the error message. This option is primarily intended for runtimes that
                contain error-handling logic that has to be run in-addition to logging and tracing the error, such as
                pytest or similar frameworks.

        Raises:
            ValidationError: If any of the inputs are not of a valid type.
            RuntimeError: If the method is called while using loguru backend without any active logger handles.
        """

        # Formats the error message. This does nt account for and is not intended to be parsed with loguru.
        formatted_message: str = self.format_message(message, loguru=False)

        # If the class is disabled, uses regular python 'raise exception' functionality.
        if not self.is_enabled:
            raise error(formatted_message)

        # If the backend is loguru, raises and catches the exception with loguru
        if self._backend == LogBackends.LOGURU and not isinstance(self._logger, NoneType):
            if not self.has_handles:
                message = (
                    f"Unable to properly log the requested error ({error}) with message {message}. The Console class "
                    f"is configured to use the loguru backend, but it does not have any handles. Call add_handles() "
                    f"method to add handles or disable() to disable Console operation."
                )
                raise RuntimeError(
                    textwrap.fill(
                        text=message,
                        width=self._line_width,
                        break_on_hyphens=self._break_on_hyphens,
                        break_long_words=self._break_long_words,
                    )
                )

            # Configures the logger to bind the proper flags to direct the message to log, terminal or nowhere
            self._logger = self._logger.bind(ataraxis_log=log, ataraxis_terminal=terminal)

            # Mostly here to prevent mypy being annoying, as the error is not really possible
            if isinstance(self._logger, NoneType):
                message = (
                    "Unable to bind the logger to use the required extra variables. Generally, this error should not "
                    "be possible."
                )
                raise RuntimeError(
                    textwrap.fill(
                        text=message,
                        width=self._line_width,
                        break_on_hyphens=self._break_on_hyphens,
                        break_long_words=self._break_long_words,
                    )
                )

            with self._logger.catch(reraise=reraise, onerror=callback):
                # noinspection PyCallingNonCallable
                raise error(formatted_message)

        # If the backend is click, prints the message to the requested destinations (file, terminal or both) and
        # optionally raises the error if re-raising is requested.
        elif self._backend == LogBackends.CLICK:
            if log:
                with open(file=str(self._error_log_path), mode="a") as file:
                    click.echo(file=file, message=formatted_message, color=False)
            if terminal:
                click.echo(message=formatted_message, err=True, color=self._use_color)
            if reraise:
                raise error(formatted_message)
